﻿using System;
using System.Collections.Generic;

namespace IMMeDotNet {
	partial class Host {

		/// <summary>
		/// Invoked when a full <see cref="IncomingUsbMessage"/> has been received.
		/// </summary>
		/// <param name="message">The <see cref="IncomingUsbMessage"/> that has been received from the wireless adaptor.</param>
		private void ProcessIncomingUsbMessage(IncomingUsbMessage message) {
			switch (message.MessageClass) {
				case 0x80:
					// Computer identification.
					var identification = new OutgoingUsbMessage(0x00, 0x01, 0x80);
					identification.Write((byte)0x00);
					identification.Write((uint)0x00000000); // Not sure what this is for.
					identification.Write(System.Net.Dns.GetHostName());
					this.Write(identification);
					break;
				case 0x00:
					if (message.ID1 == 0x01 && message.ID2 == 0x00) {

						// Add/Remove friend.
						var messageId = message.ReadByte();
						message.ReadByte();
						var contactID = message.ReadUInt32();
						bool isAdding = message.ReadByte() != 0;
						var friendName = message.ReadString();

						// Do we recognise that contact?
						Contact contact;
						if (this.connectedContacts.TryGetValue(contactID, out contact)) {


							if (isAdding) {

								// Add the friend.

								var friend = new Contact(friendName);

								Dictionary<uint, Contact> currentFriendList = null;
								if (this.connectedContactContacts.TryGetValue(contactID, out currentFriendList)) {
									currentFriendList.Add(friend.ID, friend);
								}

								// Respond to the addition request.

								var response = message.CreateResponse();
								response.Write(messageId);
								response.Write((byte)0x01);
								response.Write(contactID);
								response.Write(Error.OK);
								response.Write(friend.ID);
								this.Write(response);

								// Fire the event.

								this.OnFriendAdded(new FriendEventArgs(contact, friend));

							} else {

								// Delete the friend.

								Contact friend = null;
								uint friendID = 0;
								Dictionary<uint, Contact> currentFriendList;
								if (this.connectedContactContacts.TryGetValue(contactID, out currentFriendList)) {
									foreach (var item in currentFriendList) {
										if (item.Value.Username == friendName) {
											friend = item.Value;
											friendID = item.Key;
										}
									}
									if (currentFriendList.ContainsKey(friendID)) currentFriendList.Remove(friendID);
								}

								// Respond to the deletion request.

								var response = message.CreateResponse();
								response.Write(messageId);
								response.Write((byte)0x01);
								response.Write(contactID);
								response.Write(Error.OK);
								response.Write(friendID);
								this.Write(response);
								
								// Fire the event.

								if (friend != null) this.OnFriendRemoved(new FriendEventArgs(contact, friend));

							}
						}

					} else {
						Console.WriteLine(message);
					}
					break;
				case 0x01:
					if (message.ID1 == 0x01 && message.ID2 == 0x00) {

						// Log in attempt.

						var messageId = message.ReadByte();
						message.ReadByte();
						var messageKey = message.ReadUInt32();
						var username = message.ReadString();
						var password = message.ReadString();

						// Who is it?
						var contact = new Contact(username);

						// Attempt to log in (fire event).
						var authenticationArgs = new AuthenticatedContactEventArgs(contact, password, false);
						this.OnAttemptedLogIn(authenticationArgs);

						// Create a response.
						var loginResponse = new OutgoingUsbMessage(0x00, 0x01, 0x01);
						loginResponse.Write(messageId);
						loginResponse.Write((byte)(authenticationArgs.Success ? 0x01 : 0x00));
						loginResponse.Write((uint)(authenticationArgs.Success ? authenticationArgs.Contact.ID : 0x00000000));
						loginResponse.Write(authenticationArgs.Success ? Error.OK : Error.IncorrectPassword);
						loginResponse.Write(messageKey);
						this.Write(loginResponse);

						// Accept the connection if log in was successful.
						if (authenticationArgs.Success) {
							if (this.connectedContacts.ContainsKey(contact.ID)) {
								this.connectedContacts.Remove(contact.ID);
							}
							if (this.connectedContactContacts.ContainsKey(contact.ID)) {
								this.connectedContactContacts.Remove(contact.ID);
							}
							this.connectedContacts.Add(contact.ID, contact);
							this.OnLogIn(new ContactEventArgs(contact));
						}

					} else if (message.ID1 == 0x01 && message.ID2 == 0x08) {

						// Get friend list.
		
						var messageID = message.ReadByte();
						message.ReadByte();
						var contactID = message.ReadUInt32();

						// Do we know that contact?
						Contact contact = null;
						if (this.connectedContacts.TryGetValue(contactID, out contact)) {
							
							// Respond with a friend list.
							var friends = new FriendListEventArgs(contact);
							this.OnRequestContactList(friends);

							// Update the internal list of friends.
							Dictionary<uint, Contact> internalFriendList ;
							if (!this.connectedContactContacts.TryGetValue(contactID, out internalFriendList)) {
								this.connectedContactContacts.Add(contactID, internalFriendList = new Dictionary<uint, Contact>());
							}
							internalFriendList.Clear();
							foreach (var friend in friends.Friends) {
								internalFriendList.Add(friend.ID, friend);
							}

							// Build up the friend list response message.
							var friendList = new OutgoingUsbMessage(0x00, 0x09, 0x01);
							
							friendList.Write(messageID);
							friendList.Write((byte)0x01);
							friendList.Write(contactID);
							friendList.Write((byte)0x00);

							// Number of friends.
							friendList.Write((byte)friends.Friends.Length);
							
							// Write each friend as an ID and a username.
							foreach (var friend in friends.Friends) {
								friendList.Write(friend.ID);
								friendList.Write(friend.Username);
							}

							this.Write(friendList);

						}
					} else if (message.ID1 == 0x01 && message.ID2 == 0x04) {

						// Who's Online.

						var messageID = message.ReadByte();
						message.ReadByte();
						var contactID = message.ReadUInt32();

						// Do we know that contact?
						Contact contact;
						if (this.connectedContacts.TryGetValue(contactID, out contact)) {

							// Respond with a list of "who's online".
							var onlineFriends = new FriendListEventArgs(contact);
							this.OnRequestOnlineFriendList(onlineFriends);

							// Build up the "who's online" list response message.
							var onlineFriendList = new OutgoingUsbMessage(0x00, 0x05, 0x01);

							onlineFriendList.Write(messageID);
							onlineFriendList.Write((byte)0x01);
							onlineFriendList.Write(contactID);
							onlineFriendList.Write((byte)0x00);

							// Number of friends.
							onlineFriendList.Write((byte)onlineFriends.Friends.Length);

							// Write each friend as an ID only (no username, that was exchanged earlier).
							foreach (var friend in onlineFriends.Friends) {
								onlineFriendList.Write(friend.ID);
							}

							this.Write(onlineFriendList);
						}

					} else if (message.ID1 == 0x01 && message.ID2 == 0x02) {
						
						// Disconnection.

						var messageID = message.ReadByte();
						message.ReadByte();
						var contactID = message.ReadUInt32();
						
						// Do we know that contact?
						Contact contact;
						if (this.connectedContacts.TryGetValue(contactID, out contact)) {

							// Acknowledge disconnection.
							var disconnectionResponse = new OutgoingUsbMessage(0x00, 0x03, 0x01);

							disconnectionResponse.Write(messageID);
							disconnectionResponse.Write((byte)0x01);
							disconnectionResponse.Write(contactID);
							disconnectionResponse.Write((byte)0x00);

							this.Write(disconnectionResponse);

							// Remove the contact from our internal list.
							this.connectedContacts.Remove(contactID);
							if (this.connectedContactContacts.ContainsKey(contactID)) this.connectedContactContacts.Remove(contactID);

							this.OnLogOff(new ContactEventArgs(contact));
						}

					} else {
						Console.WriteLine(message);
					}
					break;
				case 0x02:
					
					// Text Messages.

					if (message.ID1 == 0x01 && message.ID2 == 0x04) {

						// We've received a text message.

						var messageID = message.ReadByte();
						message.ReadByte();
						var contactID = message.ReadUInt32();
						var textID = message.ReadUInt32();
						var recipientID = message.ReadUInt32();

						var text = message.ReadString();

						// Is this message deliverable to anyone?
						Contact sender = null;
						if (this.connectedContacts.TryGetValue(contactID, out sender)) {
		
							// Does the recipient know the sender?
							Contact recipient = null;
							Dictionary<uint, Contact> recipientContacts = null;
							if (this.connectedContactContacts.TryGetValue(contactID, out recipientContacts)) {
								if (!recipientContacts.TryGetValue(recipientID, out recipient)) {
									recipient = null;
								}
							}

							// Acknowledge the message.
							var textAcknowledgement = message.CreateResponse();
							textAcknowledgement.Write(messageID);
							textAcknowledgement.Write((byte)0x01);
							textAcknowledgement.Write(contactID);
							textAcknowledgement.Write(textID);
							textAcknowledgement.Write(recipient == null ? Error.NotInFriendList : Error.OK);

							this.Write(textAcknowledgement);

							// Fire an event about the received message.
							var textMessage = new MessageEventArgs(new Message(sender, recipient, text));
							if (sender != null && recipient != null) {
								this.OnMessageReceived(textMessage);
							}

						}
					} else if (message.ID1 == 0x00 && message.ID2 == 0x07) {
						
						// Acknowledgement for a message we've sent.

						var messageID = message.ReadByte();
						message.ReadByte();
						var contactID = message.ReadUInt32();
						var textID = message.ReadUInt32();
						var recipientID = message.ReadUInt32();
						message.ReadByte();

						// Try to find the message.
						Message textMessage = null;
						if (this.outbox.TryGetValue(textID, out textMessage)) {

							// Mark this as the most recently received message for the friend.
							var messageIDsToRemove = new List<uint>();
							foreach (var item in this.messageToContactMap) {
								if (item.Value.Key.ID == contactID && item.Value.Value.ID == recipientID) {
									messageIDsToRemove.Add(item.Key);
								}
							}
							foreach (var item in messageIDsToRemove) {
								this.messageToContactMap.Remove(item);
							}
							this.messageToContactMap.Add(textID, new KeyValuePair<Contact, Contact>(textMessage.Recipient, textMessage.Sender));

							this.outbox.Remove(textID);
							this.OnMessageSent(new MessageEventArgs(textMessage));
						}

					} else if (message.ID1 == 0x01 && message.ID2 == 0x08) {

						// Ending a chat.

						var messageID = message.ReadByte();
						message.ReadByte();
						var contactID = message.ReadUInt32();
						var textID = message.ReadUInt32();
						message.ReadByte(); // ?

						// We're trying to end a chat.
						Contact contact;
						KeyValuePair<Contact, Contact> friend;
						if (this.connectedContacts.TryGetValue(contactID, out contact) && this.messageToContactMap.TryGetValue(textID, out friend)) {

							// Remove the text message ID.

							this.messageToContactMap.Remove(textID);

							// Respond to the chat end event.

							var response = message.CreateResponse();
							response.Write(messageID);
							response.Write((byte)0x01);
							response.Write(contactID);
							response.Write(textID);
							response.Write(Error.OK);

							this.Write(response);

							// Fire the event.

							this.OnDiscussionEnded(new FriendEventArgs(contact, friend.Value));

						}

					} else {
						Console.WriteLine(message);
					}
					break;
				default:
					Console.WriteLine(message);
					break;
			}
		}

		private bool TryGetContactAndFriend(uint contactID, uint friendID, out Contact contact, out Contact friend) {
			if (this.connectedContacts.TryGetValue(contactID, out contact)) {
				Dictionary<uint, Contact> contactFriends;
				if (this.connectedContactContacts.TryGetValue(contactID, out contactFriends)) {
					return contactFriends.TryGetValue(friendID, out friend);
				}
			}
			friend = null;
			return false;
		}

	}
}
